from pathlib import Path
from typing import List

from src.user_errors import NoItemToRenameError
from src.user_types import Inode, InodesPaths


def paths_to_inodes_paths(paths: List[Path]) -> InodesPaths:
    """
    Given a list of paths, return a mapping from inodes to paths.

    Args:
        paths: list of Path objects

    Raises:
        FileNotFoundError: if any of the paths does not exist.
        NoItemToRenameError: empty single text file as a command-line argument.

    Returns:
        A mapping from inodes to paths.
    """
    result = {}
    missing_paths = []
    for path in paths:
        if path.exists():
            result[Inode(path.stat().st_ino)] = path
        else:
            missing_paths.append(path)
    if missing_paths:
        n = len(missing_paths)
        raise FileNotFoundError(f"{n} missing item{'s'[:n^1]}: {list(map(str,missing_paths))}.")
    elif not result:
        raise NoItemToRenameError("No item to rename was provided.")
    else:
        return result



import re
from pathlib import Path
from unicodedata import combining, normalize

LATIN = "ä  æ  ǽ  đ ð ƒ ħ ı ł ø ǿ ö  œ  ß  ŧ ü "
ASCII = "ae ae ae d d f h i l o o oe oe ss t ue"


def remove_diacritics(s: str, outliers=str.maketrans(dict(zip(LATIN.split(), ASCII.split())))):
    # more info on: https://stackoverflow.com/questions/517923/what-is-the-best-way-to-remove-accents-normalize-in-a-python-unicode-string/71408065#71408065
    return "".join(c for c in normalize("NFD", s.lower().translate(outliers)) if not combining(c))


def string_to_pairs(s, pairs=re.compile(r"(\D*)(\d*)").findall):
    # more info on: https://stackoverflow.com/questions/4836710/is-there-a-built-in-function-for-string-natural-sort/71402061#71402061
    return [(text, int(digits or 0)) for (text, digits) in pairs(s)[:-1]]


def rm_tree(path: Path):  # https://stackoverflow.com/a/57892171/173003
    if not path.is_dir():
        return
    for child in path.iterdir():
        if child.is_file():
            child.unlink()
        else:
            rm_tree(child)
    path.rmdir()



import pytest

__import__("sys").path[0:0] = "."
from src.context import Context
from src.get_editor_command import *


def test_when_favorite_editor_is_set():
    config_path = Path("test") / "workspace" / "config.json"
    config_path.write_text('{"editor_command": "mock_favorite_command"}')
    context = Context("mockOS")
    assert get_editor_command(context, Path("foobar")) == "mock_favorite_command foobar"


def test_with_unsupported_platform():
    with pytest.raises(UnsupportedOSError):
        get_editor_command(Context("MS-DOS"), Path("foobar"))


def test_with_mock_os():
    config_path = Path("test") / "workspace" / "config.json"
    if config_path.exists():
        config_path.unlink()  # Python 3.8 and newer: use `missing_ok`` parameter.
    context = Context("mockOS")
    assert get_editor_command(context, Path("foobar")) == "mock_default_command foobar"


if __name__ == "__main__":  # pragma: no cover
    pytest.main(["-qq", __import__("sys").argv[0]])



import subprocess
import sys
from argparse import ArgumentParser, RawDescriptionHelpFormatter
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import List, Dict, Any

sys.path[0:0] = ["."]

from src.file_system import FileSystem
from src.get_editable_text import get_editable_text
from src.get_editor_command import get_editor_command
from src.context import Context
from src.parse_edited_text import parse_edited_text
from src.paths_to_inodes_paths import paths_to_inodes_paths
from src.renamings import Renamer
from src.secure_clauses import secure_clauses
from src.user_errors import *
from src.user_types import EditedText


def main():
    context = Context()
    context.logger.create_new_log_file()
    context.logger.info("Starting the program.")
    context.logger.info("Parsing arguments.")
    kwargs = cli_arguments()
    context.logger.info("Parsing arguments done.")
    if kwargs["paths"]:
        do_renamings(context, **kwargs)
    else:
        undo_renamings(context)
    context.logger.info("Exiting the program.")


def undo_renamings(context: Context):
    logger = context.logger
    print_ = context.print_
    logger.info("Undoing renamings.")
    renamer = Renamer(context)
    try:
        arcs_for_undoing = renamer.get_arcs_for_undoing(logger.previous_log_contents)
        opening = "The previous renaming session was undone."
        closing = "Launch Suprenam again to restore."
        n = renamer.perform_renamings(arcs_for_undoing)
        if n == 0:
            print_.abort(f"{opening} There was no renaming to undo.")
        elif n == 1:
            print_.success(f"{opening} The sole renaming was undone. {closing}")
        else:
            print_.success(f"{opening} All {n} renamings were undone. {closing}")
    except RecoverableRenamingError:
        try:
            renamer.rollback_renamings()
            print_.abort("The renamings failed, but were successfully rolled back.")
        except Exception as e:
            return print_.fail(f"Unrecoverable failure during rollback: {e}")
    except Exception as e:  # Unknown problem with the log file, e.g. not found
        return print_.fail(f"Unrecoverable failure during undo: {e}")
    logger.info("Undoing renamings done.")


def do_renamings(context: Context, **kwargs):
    logger = context.logger
    print_ = context.print_
    logger.info("Constructing the list of items to rename.")

    paths: List[Path] = []
    if len(kwargs["paths"]) > 1:
        logger.info("The paths of several items to rename are provided.")
        paths.extend(map(Path, kwargs["paths"]))
    else:  # `do_renamings` cannot be called without at least one path. So there is exactly one.
        single_path = Path(kwargs["paths"][0])
        logger.info(f"A single path is provided: {single_path}.")
        if single_path.is_dir():
            logger.info(f"It is a directory: it and its children are to be renamed.")
            paths.append(single_path)
            paths.extend(single_path.iterdir())
        elif single_path.is_file() and single_path.suffix == ".txt":
            logger.info(f"It is a text file containing the paths of the items to rename.")
            paths.extend(map(Path, filter(None, single_path.read_text().splitlines())))
        else:
            logger.info(f"It is either a missing or a non-text file: default to rename it.")
            paths.append(single_path)
            # The case of a missing single file will be catched by `paths_to_inodes_paths()`.

    logger.info("Creating a mapping from inodes to paths.")
    try:
        inodes_paths = paths_to_inodes_paths(paths)
        logger.info("Creating a mapping from inodes to paths done.")
    except Exception as e:
        return print_.abort(str(e))

    logger.info("Creating a temporary text file for the list to be edited.")
    try:
        editable_file_path = Path(NamedTemporaryFile(mode="w+", delete=False, suffix=".tsv").name)
        logger.info(f"Editable file path: {repr(editable_file_path)}.")
    except Exception as e:
        return print_.abort(f"Failed to create a temporary file: {e}")

    logger.info("Populating the temporary text file with the list to be edited.")
    try:
        editable_file_path.write_text(get_editable_text(inodes_paths))
        logger.info(f"Editable file content: populated.")
    except Exception as e:
        return print_.abort(f"Failed to populate the temporary file: {e}")

    logger.info("Retrieving a command to edit the temporary text file.")
    try:
        editor_command = get_editor_command(context, editable_file_path)
        logger.info(f"The command is {editor_command}.")
    except Exception as e:
        return print_.abort(str(e))

    logger.info("Opening the editable text file in the editor and waiting it to be closed.")
    try:
        subprocess.run(editor_command, shell=True, check=True)
        logger.info("Command executed without process error.")
    except subprocess.CalledProcessError:
        return print_.abort(f"The command '{editor_command}' failed.")

    logger.info("Retrieving the content of the edited text file.")
    try:
        edited_text = EditedText(editable_file_path.read_text())
        logger.info("Line count in the edited text file: %s." % edited_text.count("\n"))
    except Exception as e:
        return print_.abort(f"Failed to read the edited text file: {e}")

    logger.info("Parsing the edited text into renaming clauses.")
    try:
        clauses = parse_edited_text(edited_text, inodes_paths)
        logger.info(f"Parsed edited text into {len(clauses)} clauses.")
    except Exception as e:
        return print_.abort(str(e))

    logger.info("Converting the clauses into a “safe” sequence of renamings.")
    try:
        arcs = secure_clauses(FileSystem(), clauses)
        logger.info(f"Converted clauses into {len(arcs)} arcs.")
    except Exception as e:
        return print_.abort(str(e))

    logger.info("Performing the actual renamings.")
    renamer = Renamer(context)
    try:
        n = renamer.perform_renamings(arcs)
        if n == 0:
            print_.abort(f"Nothing was changed in the name list.")
        elif n == 1:
            print_.success(f"One item was renamed.")
        else:
            print_.success(f"All {n} items were renamed.")
    except RecoverableRenamingError:
        logger.warning("Renaming performed with a recoverable error.")
        try:
            opening = "The renamings failed, but don't worry:"
            n = renamer.rollback_renamings()
            if n == 0:
                print_.abort(f"{opening} there was nothing to roll back.")
            elif n == 1:
                print_.abort(f"{opening} the only renaming was rolled back.")
            else:
                print_.abort(f"{opening} all {n} renamings were rolled back.")
        except Exception as e:
            return print_.fail(
                f"Unrecoverable failure during rollback: {e}"
                "Some files may have been renamed, some not."
                f"Please check the log file at '{logger.path}'."
            )


def cli_arguments() -> Dict[str, Any]:
    """
    CLI argument parser.

    Returns:
        A dictionary containing the parsed arguments.
    """
    parser = ArgumentParser(description="Easily rename items via your favorite text editor.")
    parser.add_argument(
        "paths",
        type=str,
        nargs="*",
        help=(
            "either several items to rename, "
            "or a text file enumerating the items to rename, "
            "or a directory whose children are to be renamed"
        ),
    )
    return vars(parser.parse_args())


if __name__ == "__main__":  # pragma: no cover
    main()



import re
from pathlib import Path
from shutil import which

from src.user_errors import *
from src.context import Context


def get_editor_command(
    context: Context,
    editable_file_path: Path,
) -> str:
    """
    Retrieve a command launching a text editor on a given text file.
    Args:
        context: all data relative to the current execution context (platform, logger, etc.).
        editable_file_path: the path to the text file to edit.
    Returns:
        A string representing the complete command to open this file in a text editor.
    Raises:
        NoEditorCommandsFileError: if `editor_commands.md` is not found.
        NoEditorError: if no command-line capable editor is installed.
        UninstalledFavoriteEditorError: if the favorite editor is not installed.
    """
    # Check whether the user has defined a favorite editor and it is installed.
    command = context.config.get("editor_command")
    if command:
        name = str(command).partition(" ")[0]  # make mypy happy
        if context.platform == "mockOS" or which(name):  # https://stackoverflow.com/a/34177358/173003
            return f"{command} {editable_file_path}"
        else:
            raise UninstalledFavoriteEditorError(
                f"The command '{name}' is not found. "
                "You can either install the corresponding editor "
                f"or modify it in '{context.workspace / 'config.json'}'."
            )

    # Retrieve a list of known editor commands.
    for editor_commands_folder in (".", "src"):
        editor_commands_path = Path(editor_commands_folder) / "editor_commands.md"
        if editor_commands_path.is_file():
            text = editor_commands_path.read_text()
            break
    else:
        raise NoEditorCommandsFileError(f"The file 'editor_commands.md' is not found.")

    # Among the commands known to work on the current platform, return the first one that is installed.
    supported_commands = r"(?m)^\|.+?\|.*?%s.*?\| *`(.+?)` *\|" % context.platform
    for command in re.findall(supported_commands, text):
        name = str(command).partition(" ")[0]  # make mypy happy
        if context.platform == "mockOS" or which(name):  # https://stackoverflow.com/a/34177358/173003
            return f"{command} {editable_file_path}"

    # If no known command is installed, raise an error.
    raise NoEditorError(f"No text editor found for the platform {context.platform}.")



import pytest

__import__("sys").path[0:0] = "."
from src.goodies import *

remove_diacritics_data = [
    ("hello, world", "hello, world"),
    ("42", "42"),
    ("你好，世界", "你好，世界"),
    (
        "Dès Noël, où un zéphyr haï me vêt de glaçons würmiens, je dîne d’exquis rôtis de bœuf au kir, à l’aÿ d’âge mûr, &cætera.",
        "des noel, ou un zephyr hai me vet de glacons wuermiens, je dine d’exquis rotis de boeuf au kir, a l’ay d’age mur, &caetera.",
    ),
    (
        "Falsches Üben von Xylophonmusik quält jeden größeren Zwerg.",
        "falsches ueben von xylophonmusik quaelt jeden groesseren zwerg.",
    ),
    (
        "Љубазни фењерџија чађавог лица хоће да ми покаже штос.",
        "љубазни фењерџија чађавог лица хоће да ми покаже штос.",
    ),
    (
        "Ljubazni fenjerdžija čađavog lica hoće da mi pokaže štos.",
        "ljubazni fenjerdzija cadavog lica hoce da mi pokaze stos.",
    ),
    (
        "Quizdeltagerne spiste jordbær med fløde, mens cirkusklovnen Walther spillede på xylofon.",
        "quizdeltagerne spiste jordbaer med flode, mens cirkusklovnen walther spillede pa xylofon.",
    ),
    (
        "Kæmi ný öxi hér ykist þjófum nú bæði víl og ádrepa.",
        "kaemi ny oexi her ykist þjofum nu baedi vil og adrepa.",
    ),
    (
        "Glāžšķūņa rūķīši dzērumā čiepj Baha koncertflīģeļu vākus.",
        "glazskuna rukisi dzeruma ciepj baha koncertfligelu vakus.",
    )
]


@pytest.mark.parametrize("text, expected", remove_diacritics_data)
def test_remove_diacritics(text, expected):
    result = remove_diacritics(text)
    print(result)
    assert result == expected


string_to_pairs_data = [
    ("", []),
    ("123", [("", 123)]),
    ("abc", [("abc", 0)]),
    ("123abc", [("", 123), ("abc", 0)]),
    ("abc123", [("abc", 123)]),
    ("123abc456", [("", 123), ("abc", 456)]),
    ("abc123efg", [("abc", 123), ("efg", 0)]),
]


@pytest.mark.parametrize("string, expected", string_to_pairs_data)
def test_string_to_pairs(string, expected):
    result = string_to_pairs(string)
    print(result)
    assert result == expected


# Some extracts from the test suite of the natsort library. Permalink:
# https://github.com/SethMMorton/natsort/blob/e3c32f5638bf3a0e9a23633495269bea0e75d379/tests/test_natsorted.py

sort_data = [
    (  # same as test_natsorted_can_sort_as_unsigned_ints_which_is_default()
        ["a50", "a51.", "a50.31", "a-50", "a50.4", "a5.034e1", "a50.300"],
        ["a5.034e1", "a50", "a50.4", "a50.31", "a50.300", "a51.", "a-50"],
    ),
    (  # same as test_natsorted_numbers_in_ascending_order()
        ["a2", "a5", "a9", "a1", "a4", "a10", "a6"],
        ["a1", "a2", "a4", "a5", "a6", "a9", "a10"],
    ),
    (  # same as test_natsorted_can_sort_as_version_numbers()
        ["1.9.9a", "1.11", "1.9.9b", "1.11.4", "1.10.1"],
        ["1.9.9a", "1.9.9b", "1.10.1", "1.11", "1.11.4"],
    ),
    (  # different from test_natsorted_handles_filesystem_paths()
        [
            "/p/Folder (10)/file.tar.gz",
            "/p/Folder (1)/file (1).tar.gz",
            "/p/Folder/file.x1.9.tar.gz",
            "/p/Folder (1)/file.tar.gz",
            "/p/Folder/file.x1.10.tar.gz",
        ],
        [
            "/p/Folder (1)/file (1).tar.gz",
            "/p/Folder (1)/file.tar.gz",
            "/p/Folder (10)/file.tar.gz",
            "/p/Folder/file.x1.9.tar.gz",
            "/p/Folder/file.x1.10.tar.gz",
        ],
    ),
    (  # same as test_natsorted_path_extensions_heuristic()
        [
            "Try.Me.Bug - 09 - One.Two.Three.[text].mkv",
            "Try.Me.Bug - 07 - One.Two.5.[text].mkv",
            "Try.Me.Bug - 08 - One.Two.Three[text].mkv",
        ],
        [
            "Try.Me.Bug - 07 - One.Two.5.[text].mkv",
            "Try.Me.Bug - 08 - One.Two.Three[text].mkv",
            "Try.Me.Bug - 09 - One.Two.Three.[text].mkv",
        ],
    ),
    (  # same as ns.DEFAULT for test_natsorted_supports_case_handling()
        ["Apple", "corn", "Corn", "Banana", "apple", "banana"],
        ["Apple", "Banana", "Corn", "apple", "banana", "corn"],
    ),
]


@pytest.mark.parametrize("given, expected", sort_data)
def test_string_to_pairs_as_sort_key(given, expected):
    result = sorted(given, key=string_to_pairs)
    print(result)
    assert result == expected


if __name__ == "__main__":  # pragma: no cover
    pytest.main(["-qq", __import__("sys").argv[0]])



import sys
from pathlib import Path


class Printer:

    def __init__(self, context):
        self.logger = context.logger
        if Path.cwd().name == "Resources": # We are in the app bundle (macOS)
            # Print uncolorized messages in Platypus' alert.
            def print_to_platypus(title, text):
                print(f"ALERT: {title}|{text}")
            self.print_success = lambda text: print_to_platypus("Success", text)
            self.print_abort = lambda text: print_to_platypus("No renamings", text)
            self.print_fail = lambda text: print_to_platypus("Fatal error", text)
            self.newline = lambda: print()
        else:
            # Print colorized messages on the standard error.
            self.OK = "\033[92m"
            self.WARNING = "\033[1m\033[38;5;166m"
            self.FAIL = "\033[1m\033[91m"
            self.RESET = "\033[0m"
            def print_to_stderr(ansi_markup, title, text):
                print(f"{ansi_markup}{title}: {text}{self.RESET}", file=sys.stderr)
            self.print_success = lambda text: print_to_stderr(self.OK, "Success", text)
            self.print_abort = lambda text: print_to_stderr(self.WARNING, "No renamings", text)
            self.print_fail = lambda text: print_to_stderr(self.FAIL, "Fatal error", text)
            self.newline = lambda: print("\r")

    def __call__(self, message: str):
        print(message)

    def success(self, message: str):
        self.logger.info(f"Success: {message}")
        self.print_success(message)

    def abort(self, message: str):
        self.logger.warning(f"Abort: {message}")
        self.print_abort(message)

    def fail(self, message: str):
        self.logger.error(f"Fail: {message}")
        self.print_fail(message)




import datetime
import logging
import os
import shutil


class Logger:
    def __init__(self, context):
        self.log_dir = context.workspace
        self.path = self.log_dir / "log.txt"
        self.logs_to_keep = context.config.get("logs_to_keep", 10)

    def create_new_log_file(self):
        """Remove all handlers associated with the root logger object and create a NEW log file."""
        self.previous_log_contents = self.get_contents() # used by `get_arcs_for_undoing()`
        self.backup_current_log_file()
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)
        logging.basicConfig(filename=self.path, filemode="w", level=logging.DEBUG)

    def backup_current_log_file(self):
        """
        Copy the current log file with a timestamp appended.
        Clean up the directory by keeping only the last self.logs_to_keep copies.
        """
        if self.logs_to_keep == 0:
            return
        try:
            modified = os.path.getmtime(self.path)
        except FileNotFoundError:
            return
        suffix = datetime.datetime.fromtimestamp(modified).strftime("%Y-%m-%d_%H-%M-%S.%f")
        destination = f"{self.path.parent}/log_{suffix}.txt"
        shutil.copy2(self.path, destination)
        for file in sorted(self.log_dir.glob("log_*.txt"), reverse=True)[self.logs_to_keep :]:
            file.unlink()

    def get_contents(self):  # pragma: no cover
        if self.path.is_file():
            return self.path.read_text().strip()
        else:
            return ""

    def warning(self, *args, **kwargs):
        logging.warning(*args, **kwargs)

    def info(self, *args, **kwargs):
        logging.info(*args, **kwargs)

    def error(self, *args, **kwargs):
        logging.error(*args, **kwargs)



from platform import platform as get_platform_long_string
from pathlib import Path
import json

from src.user_errors import *
from src.logger import Logger
from src.printer import Printer


class Context:

    DEFAULT_CONFIG = {
        "editor_command": "",
        "logs_to_keep": 10,
    }

    def __init__(self, platform_name: str = "", full=True):
        """
        Set up the platform and, if needed, create the Suprenam's data folder.

        Args:
            platform: the current OS, only provided during testing.
        Raises:
            UnsupportedOSError: if the current OS is not supported.
        """
        self.platform = platform_name or get_platform_long_string().partition("-")[0]

        if self.platform == "macOS":
            self.workspace = Path.home() / "Library" / "Application Support" / "Suprenam"
        elif self.platform == "Linux":
            self.workspace = Path.home() / ".suprenam"
        elif self.platform == "Windows":
            self.workspace = Path.home() / "AppData" / "Roaming" / "Suprenam"
        elif self.platform == "mockOS":
            self.workspace = Path("test") / "workspace"
        else:
            raise UnsupportedOSError(f"Unsupported operating system: {self.platform}.")

        self.workspace.mkdir(
            parents=True,  # any missing parents of this path are created as needed
            exist_ok=True,  # if the directory already exists, do not raise an exception
        )

        config_path = self.workspace / "config.json"
        if not config_path.is_file():
            config_path.write_text(json.dumps(self.DEFAULT_CONFIG, indent=4))
        self.config = json.loads(config_path.read_text())

        self.logger = Logger(self)

        self.print_ = Printer(self)



import re
import subprocess
from pathlib import Path
from typing import List

from src.context import Context
from src.user_errors import RecoverableRenamingError
from src.user_types import Arc


class Renamer:
    def __init__(self, context: Context, testing: bool = False):
        self.logger = context.logger
        self.print_ = context.print_
        self.rename_one_file = self._rename_one_file_with_git  # set the default renaming strategy
        if testing:
            self.logger.create_new_log_file()

    def perform_renamings(self, arcs: List[Arc]) -> int:
        """
        Perform the renamings specified by the list of arcs (source path, target path), and
        log them one by one.

        Args:
            arcs: list of couples (source_path, target_path)
        """
        n = len(arcs)
        self.print_(f"Renaming {n} items...")
        self.logger.info(f"{n} item{'s'[:n^1]} to rename.")
        try:
            self.rename_and_log_all_files(arcs)
            self.logger.info(f"{n} item{'s'[:n^1]} renamed.")
            self.print_(f"{n} item{'s'[:n^1]} renamed.")
            return n
        except Exception as e:
            self.logger.warning(f"perform_renamings: {e}")
            raise RecoverableRenamingError(f"{e}.")

    def rollback_renamings(self) -> int:
        """
        Rollback the first renaming operations.

        Note: the inverse renamings are appended to the log file.
        """
        n = len(self.arcs_to_rollback)
        self.print_(f"Rolling back the first {n} renaming{'s'[:n^1]}...")
        self.logger.info(f"{n} renaming{'s'[:n^1]} to roll back.")
        try:
            self.rename_and_log_all_files(self.arcs_to_rollback)
            self.logger.info(f"{n} renaming{'s'[:n^1]} rolled back.")
            return n
        except Exception as e:
            self.logger.error(f"rollback_renamings: {e}")
            raise

    def get_arcs_for_undoing(
        self,
        previous_log_text: str,
        get_logged_arcs=re.compile(r"(?m)^\w+:\w+:SOURCE:(.+)\tTARGET:(.+)").findall,
    ):
        """Read a log file and calculate the reversed renamings."""
        if re.search(r"(?m)^ERROR:", previous_log_text):  # The log file contains an error.
            raise ValueError("The previous rollback failed. Undoing is not possible.")
        arcs = []
        for (source, target) in reversed(get_logged_arcs(previous_log_text)):
            arcs.append(Arc(Path(target), Path(source)))
        return arcs

    def rename_and_log_all_files(self, arcs: List[Arc]):
        self.arcs_to_rollback: List[Arc] = []
        for (source, target) in arcs:
            git_flag = self.rename_one_file(source, target)
            self.logger.info(f"{'git:' if git_flag else ''}SOURCE:{source}\tTARGET:{target}")
            self.arcs_to_rollback.insert(0, Arc(target, source))
        self.print_arcs(arcs)
        self.print_.newline()

    def _rename_one_file_with_git(self, source: Path, target: Path) -> bool:
        """Try to use git to operate a renaming (default strategy for `rename_one_file()`).

        Args:
            source (Path): the path to the file to rename
            target (Path): the new path to the file

        Returns:
            bool: `True` if the renaming was done using git, `False` otherwise. In the latter case,
                the renaming is done without git, as well as all the subsequent ones.
        """
        try:
            subprocess.run(
                [
                    "git",
                    "-C",  # Run git as if it was started...
                    source.parent,  #  ... from the source's parent directory.
                    "mv",
                    source.name,
                    target.name,
                ],
                check=True,
                stderr=subprocess.DEVNULL,
            )
            return True
        except FileNotFoundError:
            self.rename_one_file = self._rename_one_file
            self.logger.warning(f"Git is not installed. Falling back to a non-git strategy.")
        except subprocess.CalledProcessError:
            pass
        except Exception as e:
            self.logger.warning(f"Error while git-renaming '{source}' to '{target}': {e}.")
        return self._rename_one_file(source, target)

    def _rename_one_file(self, source: Path, target: Path) -> bool:
        """Fallback strategy as soon as a git-renaming has found that git was not installed."""
        source.rename(target)
        return False

    def print_arcs(self, arcs: List[Arc]):
        previous_parent = Path()
        for (source, target) in arcs:
            if source.parent != previous_parent:
                self.print_.newline()
                self.print_(f"{source.parent}")
                previous_parent = source.parent
            self.print_(f"{source.name} -> {target.name}")
